// -*- Mode:C++ -*-
//Header:
//File: gcell.h
//Author: NODA Itsuki
//Date: 1998/2/8
//EndHeader:

//ModifyHistory:
// 1998/2/8: Create this file
//EndModifyHistory:

#ifndef _gcell_h_
#define _gcell_h_
////////////////////////////////////////////////////////////////////////
#ifdef NULL
#undef NULL
#endif
#define NULL 0

#include <iostream.h>
#ifdef THREADED
#include <pthread.h>
#endif

#include "thread_util.h"
#include "mmtoolT.h"
#include "ringT.h"
#include "gc.h"
#include "gdata.h"

//======================================================================
// GTermHeadArity
class GTermHeadArity {
   public:
      Symbol *head ;
      Int arity ;

      //--------------------------------------------------
      // Creater

      inline GTermHeadArity(Symbol *hd = NULLPTR, const Int a = 0) {
	 head = hd ;
	 arity = a ;
      } ;
      inline GTermHeadArity(const Term& term) {
	 head = term.head ;
	 arity = term.arity ;
      } ;
      inline GTermHeadArity(const Term* term) {
	 head = term->head ;
	 arity = term->arity ;
      } ;
      inline GTermHeadArity(const GData& gd) {
	 if(isGTerm(gd)) {
	    head = gd.val.term->head ;
	    arity = gd.val.term->arity ;
	 } else if(isGSymbol(gd)) {
	    head = gd.val.symbol ;
	    arity = 0 ;
	 } else {
	    head = NULLPTR ;
	    arity = 0 ;
	 }
      } ;
      inline GTermHeadArity(const GTermHeadArity& src) {
	 head = src.head ;
	 arity = src.arity ;
      } ; 

      //--------------------------------------------------
      // equal and let
      inline bool eq(const GTermHeadArity& target) const {
	 return (head == target.head)&&(arity == target.arity) ;
      } ;

      inline GTermHeadArity operator=(const GTermHeadArity& src) {
	 head = src.head ; arity = src.arity ;
	 return *this ;
      } ;

      //--------------------------------------------------
      // output
      inline friend ostream& operator<< (ostream& o, const GTermHeadArity& ha){
	 o << ha.head << "/" << ha.arity ;
	 return o ;
      } ;
} ;

//======================================================================
// GClause

inline bool isFactTerm(const GData &def) {
   return (isGTerm(def) 
	   && 
	   (_TermHead(def) == _SymVal(FactOpSymbol))
	   &&
	   (_TermArity(def) == 1)) ;
} ;
inline bool isRuleTerm(const GData &def) {
   return (isGTerm(def) 
	   && 
	   (_TermHead(def) == _SymVal(RuleOpSymbol))
	   &&
	   (_TermArity(def) == 2)) ;
} ;

class GClause {
   public:
      GTermHeadArity head_arity ;
      GData head ;
      GData body ;
     
      //--------------------------------------------------
      // Creater

      inline GClause() {} ;

      inline GClause(const GData &def) {
	 if (isFactTerm(def)) {
	    init(_TermArgv(def,0),TrueSymbol) ;
	 } else if (isRuleTerm(def)) {
	    init(_TermArgv(def,0),_TermArgv(def,1)) ;
	 } else {
	    cout << "Error: Illegal rule/fact: " << def << endl ;
	 } 
      } ;

      inline GClause(const GData &h, const GData &b) {
	 init(h,b) ;
      } ;

      inline bool init(const GData &h, const GData &b) {
	 head_arity = GTermHeadArity(h) ;
	 head = h ;
	 body = b ;
      } ;
	 

      //--------------------------------------------------
      // Printer

      inline friend ostream& operator<< (ostream& o, const GClause& gc) {
	 o << gc.head << " :- " << gc.body << "." ;
	 return o ;
      } ;

} ;

//======================================================================
// GClause Table

typedef Ring<GClause> GClauseRing ;
const GClause NullGClause ; 

class GClauseTable {
   public:
      GClauseRing* root ;
      GClauseRing* trash ;
      GMutex mutex ;
      
      //--------------------------------------------------
      // Creater

      inline GClauseTable() {
	 root = new GClauseRing(NullGClause,true) ;
	 trash = new GClauseRing(NullGClause,true) ;
      } ;

      //--------------------------------------------------
      // New Ring

      GClauseRing* newRing() ;
	 
      //--------------------------------------------------
      // asserta

      GClauseRing* asserta(const GClause& gclause) ;
	    
      //--------------------------------------------------
      // assertz
	    
      GClauseRing* assertz(const GClause& gclause) ;

      //--------------------------------------------------
      // retract
	    
      GClauseRing* retract(GClauseRing* gclring) ;

      //--------------------------------------------------
      // describe

      void describe(ostream& o, bool detailp = true) const ;

      //--------------------------------------------------
      // Output
      inline friend ostream& operator<< (ostream& o, 
					 const GClauseTable& gcltable) {
	 gcltable.describe(o,false) ;
	 return o ;
      } ;
} ;

//======================================================================
// CellVar Table

class CellVar {
   public:
      GData name ;
      GData value ;
      bool assigned ;
} ;

typedef Ring<CellVar> CellVarRing ;

class CellVarTable : public CellVarRing {
} ;

//======================================================================
// GCell
class GCell ;
typedef AStack<GCell*> GCellStack ;

#define GCELLSTACKSIZE	4

class GCell {
   public:
      Symbol* name ;
      GData gsymbol ;
      GClauseTable gclauseTable ;
      GCellStack gcellStack ;
      GMutex gcs_mutex ;

      //--------------------------------------------------
      // Creater

      inline GCell() {
	 init(_SymVal(NullSymbol)) ;
      } ;
      inline GCell(Symbol* nm) {
	 init(nm) ;
      } ;
      inline GCell(const GData& gd) {
	 init(gd) ;
      } ;

      inline void init(const GData gd) {
	 if(isGSymbol(gd)) {
	    init(_SymVal(gd)) ;
	 } else {
	    init(_SymVal(NullSymbol)) ;
	 }
      } ;
      inline void init(Symbol* nm) {
	 if(!(isNullPtr(nm))) {
	    name = nm ;
	    gsymbol = GSymbol(nm) ;
	    nm->gcell = this ;
	    gcellStack.init(nm->name,GCELLSTACKSIZE) ;
	 }
      } ;

      //--------------------------------------------------
      // assert/retract

      inline GClauseRing* asserta(const GClause& gclause) {
	 return gclauseTable.asserta(gclause) ;
      } ;
      inline GClauseRing* assertz(const GClause& gclause) {
	 return gclauseTable.assertz(gclause) ;
      } ;

      inline GClauseRing* retract(GClauseRing* gclring) {
	 return gclauseTable.retract(gclring) ;
      } ;

      //--------------------------------------------------
      // push/pop gcell

      inline GCellStack* pushGCell(GCell* gcell) {
	 gcs_mutex.lock() ;
	 GCellStack* r = gcellStack.push(gcell) ;
	 gcs_mutex.unlock() ;
	 return r ;
      } ;
      
      inline GCell* popGCell() {
	 gcs_mutex.lock() ;
	 GCell* r = gcellStack.pop() ;
	 gcs_mutex.unlock() ;
	 return r ;
      } ;

      inline Int removeGCell(GCell* gcell) {
	 gcs_mutex.lock() ;
	 Int r = 0 ;
	 for(UInt i = 0 ; i < gcellStack.toppos() ; i++) {
	    if(gcellStack.topN(i) == gcell) {
	       gcellStack.removeN(i) ;
	       i-- ;
	       r++ ;
	    } 
	 }
	 gcs_mutex.unlock() ;
	 return r ;
      } ;

      inline bool remove1GCell(GCell* gcell) {
	 gcs_mutex.lock() ;
	 bool r = false ;
	 for(UInt i = 0 ; i < gcellStack.toppos() ; i++) {
	    if(gcellStack.topN(i) == gcell) {
	       gcellStack.removeN(i) ;
	       r = true ;
	       break ;
	    } 
	 }
	 gcs_mutex.unlock() ;
	 return r ;
      } ;

      //--------------------------------------------------
      // copy cell stack
      UInt copyGCellStack(GCell* src) ;

      //--------------------------------------------------
      // check gcell is in gcell stack
      bool isInGCellStack(GCell* target) ;

      //--------------------------------------------------
      // describe

      void describe(ostream& o, bool detailp = true) const ;

      void showGCellStack(ostream& o, bool newlinep = false) const ;

      void listing(ostream& o, bool recursive = false, 
		   bool detailp = false) const ;

      //--------------------------------------------------
      // Output
      inline friend ostream& operator<< (ostream& o, const GCell& gcell) {
	 gcell.describe(o,false) ;
	 return o ;
      } ;

} ;

//======================================================================
// GCellPointer

class GCellPointer {
   public:
      GCellStack* gcellstack ;
      Int n ;

      //--------------------------------------------------
      // Creator

      inline GCellPointer() {gcellstack = NULLPTR ; n = 0 ; } ;
      inline GCellPointer(GCellStack* gcstk, Int i) {
	 gcellstack = gcstk ; n = i ;
      } ;
      inline GCellPointer(const GCellPointer& gcp) {
	 gcellstack = gcp.gcellstack ; n = gcp.n ;
      } ;

      //--------------------------------------------------
      // Output

      void describe(ostream& o, bool detailp = true) const ;

      inline friend ostream& operator<< (ostream& o, 
					 const GCellPointer& gcp) {
	 gcp.describe(o,false) ;
	 return o ;
      } ;

} ;

//======================================================================
// GClausePointer

typedef AStack<GCellPointer> GCellPointerStack ;

class GClausePointer {
   public:
      GClauseRing* top ;
      GCell* topcell ;
      GCellPointerStack gcpstack ;

      //--------------------------------------------------
      // Creator

      inline GClausePointer() { init0() ; } ;
      inline GClausePointer(GCell* tcell) { init0() ; init(tcell) ; } ;

      inline void init0() {
	 top = NULLPTR ;
	 topcell = NULLPTR ;
	 gcpstack.init("GClausePointer",GCELLSTACKSIZE) ;
      } ;

      inline void init() {
	 top = NULLPTR ;
	 topcell = NULLPTR ;
      } ;

      inline void init(GCell* tcell) {
	 settopcell(tcell) ;
	 gcpstack.reset() ;
      } ;

      //--------------------------------------------------
      // set top cell

      inline void settopcell(GCell* tcell) {
	 topcell = tcell ;
	 top = tcell->gclauseTable.root ;
      } ;

      //--------------------------------------------------
      // push/pop

      void pushgcp(GCell* tcell) ;

      GCell* popgcell() ;

      //--------------------------------------------------
      // find GClause in Ring

      GClauseRing* findGClauseInRing(const GTermHeadArity& gtermha) ;

      //--------------------------------------------------
      // find GClause

      GClauseRing* findGClause(const GTermHeadArity& gtermha) ;

      //--------------------------------------------------
      // Output

      void describe(ostream& o, bool detailp = true) const ;

      inline friend ostream& operator<< (ostream& o, 
					 const GClausePointer& gcp) {
	 gcp.describe(o,false) ;
	 return o ;
      } ;

} ;
      


////////////////////////////////////////////////////////////////////////
#endif

