// -*- Mode:C++ -*-
//Header:
//File: gdata.cc
//Author: NODA Itsuki
//Date: 1997/9/12
//EndHeader:

//ModifyHistory:
// 1997/9/12: Create this file
//EndModifyHistory:

//======================================================================
// Include Files
//
#include "gdata.h"

//======================================================================
// Symbol Table & Term Table

char * GDTypeName[] = GDTypeNameString ;

SymbolTable SymTab ;

Symbol * SymbolTable::newSymbol(const String &nm) {
   if(n < size) {
      Symbol *s = entry(n) ; n++ ;
      s->name = nm ;
      s->id = idmax ; idmax++ ;  
      return s ;
   } else if (expand()) {
      return newSymbol(nm) ;
   } else {
      cerr << "Warning: no more symbol (in generating '" << nm << "'.)" 
	   << endl ;
   } ;
   return NULL ;
} ;

//--------------------------------------------------
// gensym

Symbol * SymbolTable::genSymbol(const char* seed) {
   UInt len = strlen(seed) + 64 ;
   char* name = new char[len] ;
   for(UInt i = 0 ; true ; i++) {
      ostrstream ostr(name,len) ;
      ostr << seed << VARCONCATCHAR << i << ends ;
      Symbol* symp = _findSymbol(name) ;
      if(symp == NULLPTR) break ;
   }
   char* newname = strdup(name) ;
   delete name ;
   return findSymbol(newname) ;
} ;

//--------------------------------------------------
// Expand Table

bool SymbolTable::expand(UInt addsize) {
   Symbol ** oldentry = _entry ;
   UInt oldsize = size ;
   size = size + addsize ;
  
#ifdef DEBUG0
   cerr << "Status: expanding Symbol Table:" << oldsize << "->" << size 
	<< endl ;
#endif

   _entry = new (Symbol*)[size] ;
   if(_entry == NULL) return false ;
  
   for(UInt i = 0 ; i < oldsize ; i++) { _entry[i] = oldentry[i] ; } ;
   if(oldsize > 0) delete oldentry ;
  
   Symbol * newentry = new Symbol[addsize] ;
   if(newentry == NULL) return false ; 
    
   for(UInt i = oldsize ; i < size ; i++) {
      _entry[i] = &(newentry[i - oldsize]) ; 
   } ;
   return true ;
} ;

TermTable TermTab ;

bool TermTable::expand(UInt addsize) {
   Term ** oldentry = _entry ;
   UInt oldsize = size ;
   size = size + addsize ;
 
#ifdef DEBUG0
   cerr << "Status: expanding Term Table:" << oldsize << "->" << size 
	<< endl ;
#endif

   _entry = new (Term *)[size] ;
   if(_entry == NULL) return false ;

   for(UInt i = 0 ; i < oldsize ; i++) { _entry[i] = oldentry[i] ; } ;
   if(oldsize > 0) delete oldentry ;
    
   Term * newentry = new Term[addsize] ;
   if(newentry == NULL) return false ;
    
   for(UInt i = oldsize ; i < size ; i++) {
      _entry[i] = &(newentry[i - oldsize]) ;
   } ;
   return true ;
} ;

//--------------------------------------------------
// describe Table

void SymbolTable::describe(ostream& o, bool detailp) const {
   o << "#SymbolTable[" << n << "/" << size << "(" << idmax << ")]" ;
   if(detailp) {
      o << endl ;
      for(UInt i = 0 ; i < n ; i++) {
	 o << "\t" << *entry(i) << "\t" ;
	 if(entry(i)->funcp) o << " func" ;
	 if(!isNullPtr(entry(i)->gcell)) o << " gcell" ;
	 if(!isNullPtr(entry(i)->gem)) o << " gem" ;
	 if(!isNullPtr(entry(i)->gdbox)) o << " gdbox" ;
	 if(!isNullPtr(entry(i)->mutex)) o << " mutex" ;
	 o << endl ;
      }
   }
}

void showSymTab() { SymTab.describe(cout,true) ; } ; 

//======================================================================
// Describe Data

void Symbol::_describe(ostream& o, const bool newlinep, const bool detailp,
		       const bool quotep) 
const 
{
   if(detailp) o << "#Sym[" << id << ":" ;

   if(quotep)  o << "'" << name << "'" ;
   else        o << name ;

   if(detailp) {
      if(funcp) o << ", function=" << func ;
      if(!isNullPtr(gcell)) o << ", gcell=" << *gcell ;
      if(detailp) o << "]" ;
   }

   if(newlinep) o << endl ;
} ;
     
void GData::_describe(ostream& o, const bool newlinep, const bool detailp) 
const 
{
   switch(type) {
      case GDT_Variable: 
	 ((GVariable*)this)->_describe(o,newlinep,detailp); break;
      case GDT_Integer:  
	 ((GInteger*)this)->_describe(o,newlinep,detailp); break;
      case GDT_Float: 
	 ((GFloat*)this)->_describe(o,newlinep,detailp); break;
      case GDT_Symbol: 
	 ((GSymbol*)this)->_describe(o,newlinep,detailp); break;
      case GDT_String: 
	 ((GString*)this)->_describe(o,newlinep,detailp); break;
      case GDT_Pointer: 
	 ((GPointer*)this)->_describe(o,newlinep,detailp); break;
      case GDT_Term: 
	 ((GTerm*)this)->_describe(o,newlinep,detailp); break;
     defalt:
	 if(detailp)
	    o << "#DT[type=" << GDTypeName[type] << "]" ;
	 else
	    o << "#DT[" << this << "]" ;
	 if(newlinep) o << endl ;
   }
} ;

void Term::_describe(ostream& o, const bool newlinep, const bool detailp) 
const 
{
   if(detailp)
      o << "#Term[" << head->name << "/" << arity << "(" ;
   else
      o << head->name << "(" ;

   for(Int i = 0 ; i < arity ; i++) {
      if(i>0) o << "," ;
      (argv(i))._describe(o, false, detailp) ;
   }

   if(detailp)
      o << ")]" ;
   else 
      o << ")" ;

   if(newlinep) o << endl ;
} ;

void GInteger::_describe(ostream& o, const bool newlinep, const bool detailp) 
const 
{
   if(detailp) 
      o << "#Int[" << val.inum << "]" ;
   else
      o << val.inum ;
   if(newlinep) o << endl ;
} ;

void GFloat::_describe(ostream& o, const bool newlinep, const bool detailp) 
const 
{
   if(detailp)
      o << "#Flt[" << val.fnum << "]" ;
   else
      o << val.fnum ;
   if(newlinep) o << endl ;
} ;

void GSymbol::_describe(ostream& o, const bool newlinep, const bool detailp) 
const 
{
   (*(val.symbol))._describe(o,newlinep,detailp,shouldquotep()) ;
} ;

void GVariable::_describe(ostream& o, const bool newlinep, const bool detailp) 
const 
{
   if(detailp)
      o << "#Var" ;
  
   (*(val.symbol))._describe(o,newlinep,detailp,false) ;
} ;

void GString::_describe(ostream& o, const bool newlinep, const bool detailp) 
const 
{
   if(detailp)
      o << "#Str[" << val.string << ":\"" << *val.string << "\"]" ;
   else
      o << "\"" << *val.string << "\"" ;
   if(newlinep) o << endl ;
} ;

void GPointer::_describe(ostream& o, const bool newlinep, const bool detailp) 
const 
{
   if(detailp)
      o << "#Ptr[" << val.pointer << "]" ;
   else
      o << "#Ptr[" << val.pointer << "]" ;
   if(newlinep) o << endl ;
} ;

void GTerm::_describe(ostream& o, const bool newlinep, const bool detailp)
const 
{
   (*(val.term))._describe(o,newlinep,detailp) ;
} ;

//======================================================================
// Equal Data
bool GData::eq(const GData & gd) const 
{
   if(type == gd.type) {
      if        (type == GDT_Integer) {
	 return val.inum == gd.val.inum ;
      } else if (type == GDT_Float) {
	 return val.fnum == gd.val.fnum ;
      } else {
	 return val.pointer == gd.val.pointer ;
      }
   } else {
      return false ;
   } 
} 

//======================================================================
// Generate GSymbol and GVariable 

GData genGSymbol(const GData &seed) {
   Symbol* sym ;
   if(isGSymbol(seed)){
      sym = GenSymbol(seed.val.symbol) ;
   } else {
      sym = GenSymbol(NullSymbol.val.symbol) ;
   }
   return GSymbol(sym) ;
}

GData genGVariable(const GData &seed) {
   Symbol* sym ;
   if(isGVariable(seed)) {
      sym = GenSymbol(seed.val.symbol) ;
   } else {
      sym = GenSymbol(AnonymousVar.val.symbol) ;
   }
   return GVariable(sym) ;
} ;

//======================================================================
// Assign GData Box to Symbol

bool Symbol::assignGDBox() {
   if(hasGDBox()) return false ;
   gdbox = new GDataBox() ;
   return true ;
}

bool GDataBox::put(const GData& v, GDataBox::PutType puttype) {
   bool r = true ;
   mutex.lock() ;
   switch(puttype) {
      case PT_force: 
	 break ;
      case PT_wait: 
	 while(assigned) 
	    mutex.wait() ;
	 break ;
      case PT_nowait: 
	 if(assigned) r = false ;
	 break ;
      default: 
	 r = false ;
	 break ;
   }
   if(r) {
      val = v ;
      assigned = true ;
      mutex.signal() ;
   }
   mutex.unlock() ;
   return r ;
}

bool GDataBox::get(GData& v, bool retractp, bool waitp) {
   bool r = true ;
	 
   mutex.lock() ;
   if(waitp) {
      while(!assigned) 
	 mutex.wait() ;
   } else {
      if(!assigned) r = false ;
   }
   if(r) {
      v = val ;
      if(retractp) {
	 assigned = false ;
	 mutex.signal() ;
      }
   }
   mutex.unlock() ;
   return r ;
}



//======================================================================
// Initialize Global Data

#include "gglobal.cc" 




