// -*- Mode:C++ -*-
//Header:
//File: gdata.h
//Author: NODA Itsuki
//Date: 1997/9/9
//EndHeader:

//ModifyHistory:
// 1997/9/9: Create this file
//EndModifyHistory:

#ifndef _gdata_h_
#define _gdata_h_
////////////////////////////////////////////////////////////////////////

//======================================================================
// Include Files
//
#include <string.h>
#include <stdio.h>
#include <iostream.h>
#include <strstream.h>

#include "base.h"
#include "thread_util.h"
#include "gc.h"
#include "astableT.h"

//======================================================================
// Table Default Size
//
#define SymbolTableDefaultSize 1024
#define TermTableDefaultSize 1024

//======================================================================
// Value Type
//
enum GDType {
   GDT_None,
   GDT_Variable,
   GDT_Integer,
   GDT_Float,
   GDT_Symbol,
   GDT_Term,
   GDT_String,
   GDT_Pointer,
} ;

#define GDTypeNameString {"none",\
   "variable",\
   "integer",\
   "float",\
   "symbol",\
   "term",\
   "string",\
   "pointer",\
   "unknown" } 

extern char * GDTypeName[] ;

class Symbol ;
class Term ;
class GData ;

union GDValue {
      Int		inum ;
      Float 	fnum ;
      Symbol *	symbol ;
      Term *	term ;
      String *	string ;
      void *	pointer ;
} ;

inline bool operator==(const GDValue& a,  const GDValue& b) {
   return !strncmp((char *)&a, (char *)&b, sizeof(GDValue)) ;
} ;

//======================================================================
// Symbol
//
class GCell ;
class Gem ;
class GDataBox ;

class Symbol : public GCEntity {
   public:
      String name ;
      ID id ;
      bool funcp ;
      bool (*func)(Gem*) ;

      GCell* gcell ;
      Gem* gem ;
      GDataBox* gdbox ;
      GMutex* mutex ; 
      

      //--------------------------------------------------
      // Creater

      inline Symbol() {
	 funcp = false ; gcell = NULLPTR ; 
	 gem = NULLPTR ; gdbox = NULLPTR ; mutex = NULLPTR ;
      } ;
      inline Symbol(const String &nm, const ID i = 0) { 
	 name = nm ; id = i ; funcp = false ; gcell = NULLPTR ;
	 gem = NULLPTR ; gdbox = NULLPTR ; mutex = NULLPTR ;
      } ;
      inline Symbol(const char* nm, const ID i = 0) { 
	 name = nm ; id = i ; funcp = false ; gcell = NULLPTR ;
	 gem = NULLPTR ; gdbox = NULLPTR ; mutex = NULLPTR ;
      } ;
  
      //--------------------------------------------------
      // Equal

      inline bool eq(Symbol* that) { return this==that ; } ; 

      //--------------------------------------------------
      // set func
      
      inline void setfunc(bool (*f)(Gem*)) {
	 funcp = true ;
	 func = f ;
      } ;

      //--------------------------------------------------
      // assign gdata box

      bool assignGDBox() ;
      
      //--------------------------------------------------
      // check type

      inline bool hasFunc() const { return funcp ; } ;
      inline bool hasGCell() const { return !isNullPtr(gcell) ; } ;
      inline bool hasGem() const { return !isNullPtr(gcell) ; } ;
      inline bool hasGDBox() const { return !isNullPtr(gdbox) ; } ;
      inline bool hasMutex() const { return !isNullPtr(mutex) ; } ;

      //--------------------------------------------------
      // Printer
  
      inline void describe(ostream& o = cout, 
			   const bool newlinep = true, 
			   const bool detailp = true,
			   const bool quotep = false) const {
			      _describe(o,newlinep,detailp,quotep) ;} ;
      void _describe(ostream& o, const bool newlinep, const bool detailp,
		     const bool quotep ) const ;

      inline friend ostream& operator<< (ostream& o, const Symbol& s) {
	 s.describe(o,false,false) ;
	 return o ;
      } ;
} ;

//======================================================================
// Symbol Table
//

class SymbolTable : public GCEntityTable {
   public:
      UInt size ;
      UInt n ;
      UInt idmax ;
      Symbol ** _entry;
  
      //--------------------------------------------------
      // Creater

      inline SymbolTable(const UInt sz = SymbolTableDefaultSize) {
	 size = 0 ;
	 n = 0 ;
	 idmax = 0 ;
	 _entry = NULLPTR ;
	 Lock() ;
	 expand(sz) ;
	 UnLock() ;
      } ;

      //--------------------------------------------------
      // Get Nth Entry

      inline Symbol * entry(UInt i) const { return _entry[i] ; } ;

      //--------------------------------------------------
      // New Symbol

      inline Symbol * newSymbol(const char * nm) { 
	 return newSymbol(String(nm)) ; } ;
      Symbol * newSymbol(const String &nm) ;

      //--------------------------------------------------
      // Find Symbol

      inline Symbol * _findSymbol(const char* sym) const {
	 return _findSymbol(String(sym)) ;
      } ;

      inline Symbol * _findSymbol(const String & sym) const {
	 for(UInt i = 0 ; i < n ; i++) {
	    Symbol *e = entry(i) ;
	    if(e->name == sym) return e ;
	 }
	 return (Symbol *)NULLPTR ;
      } ;

      inline Symbol * findSymbol(const char * nm) { 
	 return findSymbol(String(nm)) ; } ;
      inline Symbol * findSymbol(const String & nm) {
	 Lock() ;
	 Symbol * s = _findSymbol(nm) ;
	 if(s == NULLPTR) { s = newSymbol(nm) ; } 
	 UnLock() ;
	 return s ;
      } ;

      //--------------------------------------------------
      // Generate New Symbol

      Symbol * genSymbol(const char* seed) ;
      inline Symbol * genSymbol(const String& seed) {
//       return genSymbol((char*)seed) ;
	 return genSymbol(seed.data) ;
      } ;
      inline Symbol * genSymbol(const Symbol* seed) {
	 return genSymbol(seed->name) ;
      } ;
       

      //--------------------------------------------------
      // Expand Table

      bool expand(UInt addsize = SymbolTableDefaultSize) ;

      //--------------------------------------------------
      // describe/output

      void describe(ostream& o, bool detailp = true) const ;

      inline friend ostream& operator<< (ostream& o, const SymbolTable& st) {
	 st.describe(o,false) ;
	 return o ;
      } ;

} ;

extern SymbolTable SymTab ;

inline Symbol * FindSymbol(const char * nm) { 
   return SymTab.findSymbol(String(nm)) ; 
} ;
inline Symbol * FindSymbol(const String & nm) {
   return SymTab.findSymbol(nm) ; 
} ;


inline Symbol * GenSymbol(const char* seed) {
   return SymTab.genSymbol(seed) ;
} ;
inline Symbol * GenSymbol(const String& seed) {
   return SymTab.genSymbol(seed) ;
} ;
inline Symbol * GenSymbol(const Symbol* seed) {
   return SymTab.genSymbol(seed) ;
} ;

extern void showSymTab() ;

//======================================================================
// Data
//
class GData : public GCNode {
   public:
      GDType type ;
      GDValue val ;

      inline void describe(ostream& o = cout, 
			   const bool newlinep = true,
			   const bool detailp = true) const {
			      _describe(o,newlinep,detailp) ;
      } ;

      void _describe(ostream& o,const bool newlinep,const bool detailp) 
   const ;

      /* followings do not work well */
//  virtual void __describe(ostream& o,const bool newlinep,const bool detailp) 
//    const ;
  
      inline friend ostream& operator<< (ostream& o, const GData& dt) {
	 dt.describe(o,false,false) ;
	 return o ;
      } ;

      bool eq(const GData &gd) const ;
      /* maybe bug 
	 inline GData operator=(const GData& dt) { 
	 _copy_data(dt) ; 
	 return *this ; 
	 } ;
	 */
      inline void _copy_data(const GData& dt) { 
	 memcpy(this,&dt,sizeof(*this)) ;
      };

      inline void init() ;
} ;

extern GData NullSymbol ;

//======================================================================
// Term
//
const Int TermArityStep = 3 ;

class Term : public GCEntity {
   public:
      Symbol * 	head ;
      Int	arity ;
      GData	_argv[TermArityStep] ;
      Term *	next ;
  
//--------------------------------------------------
// Constructor

      inline Term() {} ;

      inline Term(Symbol* hd, const GData& arg1) { 
	 head = hd ; arity = 1 ; next = NULLPTR ;
	 argv(0) = arg1 ; } ;
      inline Term(Symbol* hd, 
		  const GData& arg1, const GData& arg2) {
		     head = hd ; arity = 2 ; next = NULLPTR ;
		     argv(0) = arg1 ; argv(1) = arg2 ; } ;
      inline Term(Symbol* hd, 
		  const GData& arg1, const GData& arg2, const GData& arg3) {
		     head = hd ; arity = 3 ; next = NULLPTR ;
		     argv(0) = arg1 ; argv(1) = arg2 ; argv(2) = arg3 ; } ;
      inline Term(Symbol* hd, 
		  const GData& arg1, const GData& arg2, const GData& arg3,
		  const GData& arg4) {
		     head = hd ; arity = 4 ; next = NULLPTR ;
		     argv(0) = arg1 ; argv(1) = arg2 ; argv(2) = arg3 ; 
		     argv(3) = arg4 ; } ; 
  
//--------------------------------------------------
// Printer

      inline void describe(ostream& o = cout, 
			   const bool newlinep = true,
			   const bool detailp = true) const {
			      _describe(o,newlinep,detailp) ;
      } ;			 
			 
      void _describe(ostream& o,const bool newlinep,const bool detailp) const ;
      inline friend ostream& operator<< (ostream& o, Term& tm) {
	 tm.describe(o,false,false) ;
	 return o ;
      } ;
  
//--------------------------------------------------
// Get Arg Entry

      inline GData& argv(Int idx) const {
	 Int n = idx/TermArityStep ;
	 Term *tm = (Term *)this ;
	 for(Int i = 0 ; i < n ; i++) { tm = tm->next ; }
	 return tm->_argv[idx % TermArityStep] ;
      } ;
  
} ;

//======================================================================
// Term Table
//

class TermTable : public GCEntityTable {
   public:
      UInt size ;
      UInt n ;
      Term ** _entry ;

      //--------------------------------------------------
      // Creater

      inline TermTable(const UInt sz = TermTableDefaultSize) {
	 size = 0 ;
	 n = 0 ;
	 Lock() ;
	 expand(sz) ;
	 UnLock() ;
      } ;

      //--------------------------------------------------
      // Get Nth Entry

      inline Term * entry(UInt i) const { return _entry[i] ; } ;

      //--------------------------------------------------
      // New Term

      inline Term * _newTerm() {
	 if(n < size) { return entry(n++) ; } 
	 else if(expand()) { return _newTerm() ; }
	 else { 
	    cerr << "Warning: no more term." << endl ;
	 } ;
	 return NULLPTR ;
      } ;
      inline Term * newTerm() {
	 Lock() ;
	 Term * r = _newTerm() ;
	 UnLock() ;
	 return r ;
      } ;

      inline Term * newTerm(const UInt arity) {
	 UInt l = arity / TermArityStep ;
	 UInt r = arity % TermArityStep ;
	 if (r) l++ ;
	 Term * hd = newTerm() ;
	 hd->arity = arity ;
	 hd->next = NULLPTR ;
	 Term * md = hd ;
	 for(l-- ; l > 0 ; l--) {
	    Term * tl = newTerm() ;
	    md->next = tl ;
	    tl->next = NULLPTR ;
	    md = tl ;
	 } 
	 return hd ;
      } ;

      inline Term * newTerm(Symbol * head, const UInt arity) {
	 Term * headterm = newTerm(arity) ;
	 headterm->head = head ;
	 return headterm ;
      } ;

      //--------------------------------------------------
      // Expand Table

      bool expand(UInt addsize = TermTableDefaultSize) ;
} ;

extern TermTable TermTab ;

inline Term * NewTerm(const UInt arity) { return TermTab.newTerm(arity) ; } ;
inline Term * NewTerm(Symbol * head, const UInt arity) {
   return TermTab.newTerm(head, arity) ;
} ;
inline Term * NewTerm(Symbol *head, const GData & arg0) {
   Term * term = NewTerm(head,1) ;
   term->argv(0) = arg0 ;
   return term ;
} ;
inline Term * NewTerm(Symbol *head, const GData & arg0, const GData & arg1) {
   Term * term = NewTerm(head,2) ;
   term->argv(0) = arg0 ; term->argv(1) = arg1 ;
   return term ;
} ;
inline Term * NewTerm(Symbol *head, const GData & arg0, const GData & arg1, 
		      const GData & arg2) {
   Term * term = NewTerm(head,3) ;
   term->argv(0) = arg0 ; term->argv(1) = arg1 ; term->argv(2) = arg2 ;
   return term ;
} ;
inline Term * NewTerm(Symbol *head, const GData & arg0, const GData & arg1, 
		      const GData & arg2, const GData & arg3) {
			 Term * term = NewTerm(head,4) ;
			 term->argv(0) = arg0 ; term->argv(1) = arg1 ; 
			 term->argv(2) = arg2 ; term->argv(3) = arg3 ;
			 return term ;
} ;
inline Term * NewTerm(Symbol *head, const GData & arg0, const GData & arg1, 
		      const GData & arg2, const GData & arg3, 
		      const GData & arg4) {
			 Term * term = NewTerm(head,5) ;
			 term->argv(0) = arg0 ; term->argv(1) = arg1 ; 
			 term->argv(2) = arg2 ; term->argv(3) = arg3 ; 
			 term->argv(4) = arg4 ;
			 return term ;
} ;
inline Term * NewTerm(Symbol *head, const GData & arg0, const GData & arg1, 
		      const GData & arg2, const GData & arg3, 
		      const GData & arg4, const GData & arg5) {
			 Term * term = NewTerm(head,6) ;
			 term->argv(0) = arg0 ; term->argv(1) = arg1 ; 
			 term->argv(2) = arg2 ; term->argv(3) = arg3 ; 
			 term->argv(4) = arg4 ; term->argv(5) = arg5 ;
			 return term ;
} ;
inline Term * NewTerm(Symbol *head, const GData & arg0, const GData & arg1, 
		      const GData & arg2, const GData & arg3, 
		      const GData & arg4, const GData & arg5, 
		      const GData & arg6) {
			 Term * term = NewTerm(head,7) ;
			 term->argv(0) = arg0 ; term->argv(1) = arg1 ; 
			 term->argv(2) = arg2 ; term->argv(3) = arg3 ; 
			 term->argv(4) = arg4 ; term->argv(5) = arg5 ;
			 term->argv(6) = arg6 ;
			 return term ;
} ;
inline Term * NewTerm(Symbol *head, const GData & arg0, const GData & arg1,
		      const GData & arg2, const GData & arg3, 
		      const GData & arg4, const GData & arg5, 
		      const GData & arg6, const GData & arg7) {
			 Term * term = NewTerm(head,8) ;
			 term->argv(0) = arg0 ; term->argv(1) = arg1 ; 
			 term->argv(2) = arg2 ; term->argv(3) = arg3 ; 
			 term->argv(4) = arg4 ; term->argv(5) = arg5 ;
			 term->argv(6) = arg6 ; term->argv(7) = arg7 ;
			 return term ;
} ;


//======================================================================
// GInteger
//
class GInteger : public GData {
   public:
      inline GInteger(const Int i) {
	 type = GDT_Integer ;
	 val.inum = i ;
   } ;
  
      void _describe(ostream& o,const bool newlinep,const bool detailp) 
   const ;
//  virtual void __describe(ostream& o,const bool newlinep,const bool detailp) 
//    const ;
} ;


//======================================================================
// GFloat
//
class GFloat : public GData {
   public:
      inline GFloat(const Float & f) {
	 type = GDT_Float ;
	 val.fnum = f ;
   } ;
      void _describe(ostream& o,const bool newlinep,const bool detailp) 
   const ;
//  virtual void __describe(ostream& o,const bool newlinep,const bool detailp) 
//    const ;
} ;

//======================================================================
// GSymbol
//

class GSymbol : public GData {
   public:
      inline GSymbol(const char * nm) {
	 type = GDT_Symbol ;
	 val.symbol = FindSymbol(nm) ;
   } ;
      inline GSymbol(const String & nm) {
	 type = GDT_Symbol ;
	 val.symbol = FindSymbol(nm) ;
      } ;
      inline GSymbol(Symbol * sym = NULLPTR) {
	 type = GDT_Symbol ;
	 val.symbol = sym ;
      } ;
    
      void _describe(ostream& o,const bool newlinep,const bool detailp) 
   const ;
//  virtual void __describe(ostream& o,const bool newlinep,const bool detailp) 
//    const ;

      bool shouldquotep() const ;

} ;

//======================================================================
// GVariable
//
class GVariable : public GSymbol {
   public:
      inline GVariable(const char * nm) : GSymbol(nm) {
	 type = GDT_Variable ; 
      } ;
      inline GVariable(const String & nm) : GSymbol(nm) { 
	 type = GDT_Variable ; 
      } ;
      inline GVariable(Symbol * sym = NULLPTR) : GSymbol(sym) { 
	 type = GDT_Variable ; 
      } ; 

      void _describe(ostream& o,const bool newlinep,const bool detailp) 
   const ;
//  virtual void __describe(ostream& o,const bool newlinep,const bool detailp) 
//    const ;
} ;

//======================================================================
// GString
//
class GString : public GData {
   public:
      inline GString() {
	 type = GDT_String ;
	 val.string = new String() ;
      } ;
      inline GString(const char *s) {
	 type = GDT_String ;
	 val.string = new String(s) ;
      } ;
      inline GString(const String& s) {
	 type = GDT_String ;
	 val.string = new String(s) ;
      } ;

      inline void _let(char *s) {
	 val.string = new String() ;
	 val.string->_let(s) ;
      } ;
      inline void _let(String* s) {
	 delete val.string ;
	 val.string = s ;
      } ;

      void _describe(ostream& o, const bool newlinep, const bool detailp) const ;
} ;


//======================================================================
// GPointer
//
class GPointer : public GData {
   public:
      inline GPointer() {
	 type = GDT_Pointer ;
	 val.pointer = NULLPTR ;
      } ;
      inline GPointer(void* ptr) {
	 type = GDT_Pointer ;
	 val.pointer = ptr ;
      } ;

      inline void _let(void* ptr) {
	 val.pointer = ptr ;
      } ;
      void _describe(ostream& o, const bool newlinep, const bool detailp) const ;
} ;


//======================================================================
// GTerm
//
class GTerm : public GData {
   public:
      //------------------------------
      // constructor by symbol head
      inline GTerm() {} ;
      inline GTerm(Term * term) { type = GDT_Term ; val.term = term ; } ;
      inline GTerm(Symbol * sym, const UInt arity) {
	 type = GDT_Term ;
	 val.term = NewTerm(sym, arity) ;
      } ;
      inline GTerm(Symbol * sym, const GData & arg0) {
	 type = GDT_Term ;
	 val.term = NewTerm(sym, arg0) ;
      } ;
      inline GTerm(Symbol * sym, const GData & arg0, const GData & arg1) {
	 type = GDT_Term ;
	 val.term = NewTerm(sym, arg0, arg1) ;
      } ;
      inline GTerm(Symbol * sym, const GData & arg0, const GData & arg1, 
		   const GData & arg2) {
	 type = GDT_Term ;
	 val.term = NewTerm(sym, arg0, arg1, arg2) ;
      } ;
      inline GTerm(Symbol * sym, const GData & arg0, const GData & arg1, 
		   const GData & arg2, const GData & arg3) {
		      type = GDT_Term ;
		      val.term = NewTerm(sym, arg0, arg1, arg2, arg3) ;
      } ;
      inline GTerm(Symbol * sym, const GData & arg0, const GData & arg1, 
		   const GData & arg2, const GData & arg3, const GData & arg4){
		      type = GDT_Term ;
		      val.term = NewTerm(sym, arg0, arg1, arg2, arg3, arg4) ;
      } ;
      inline GTerm(Symbol * sym, const GData & arg0, const GData & arg1, 
		   const GData & arg2, const GData & arg3, const GData & arg4,
		   const GData & arg5) {
		      type = GDT_Term ;
		      val.term = NewTerm(sym, arg0, arg1, arg2, arg3, arg4, 
					 arg5) ;
      } ;
      inline GTerm(Symbol * sym, const GData & arg0, const GData & arg1, 
		   const GData & arg2,const GData & arg3, const GData & arg4, 
		   const GData & arg5, const GData & arg6) {
		      type = GDT_Term ;
		      val.term = NewTerm(sym, arg0, arg1, arg2, arg3, arg4, 
					 arg5, arg6) ;
      } ;
      inline GTerm(Symbol * sym, const GData & arg0, const GData & arg1, 
		   const GData & arg2, const GData & arg3, const GData & arg4,
		   const GData & arg5, const GData & arg6, const GData & arg7){
		      type = GDT_Term ;
		      val.term = NewTerm(sym, arg0, arg1, arg2, arg3, arg4, 
					 arg5, arg6, arg7) ;
      } ;

      //------------------------------
      // constructor by gdata head

      inline Symbol* symval(const GData& sym) {
	 if(sym.type == GDT_Symbol) return sym.val.symbol ;
	 else return NullSymbol.val.symbol ;
      } ;

      inline GTerm(const GData& sym, const GData & arg0) {
	 type = GDT_Term ;
	 val.term = NewTerm(symval(sym), arg0) ;
      } ;
      inline GTerm(const GData& sym, const GData & arg0, const GData & arg1) {
	 type = GDT_Term ;
	 val.term = NewTerm(symval(sym), arg0, arg1) ;
      } ;
      inline GTerm(const GData& sym, const GData & arg0, const GData & arg1, 
		   const GData & arg2) {
	 type = GDT_Term ;
	 val.term = NewTerm(symval(sym), arg0, arg1, arg2) ;
      } ;
      inline GTerm(const GData& sym, const GData & arg0, const GData & arg1, 
		   const GData & arg2, const GData & arg3) {
		      type = GDT_Term ;
		      val.term = NewTerm(symval(sym), arg0, arg1, arg2, arg3) ;
      } ;
      inline GTerm(const GData& sym, const GData & arg0, const GData & arg1, 
		   const GData & arg2, const GData & arg3, const GData & arg4){
		      type = GDT_Term ;
		      val.term = NewTerm(symval(sym), arg0, arg1, arg2, arg3, 
					 arg4) ;
      } ;
      inline GTerm(const GData& sym, const GData & arg0, const GData & arg1, 
		   const GData & arg2, const GData & arg3, const GData & arg4,
		   const GData & arg5) {
		      type = GDT_Term ;
		      val.term = NewTerm(symval(sym), arg0, arg1, arg2, arg3, 
					 arg4, arg5) ;
      } ;
      inline GTerm(const GData& sym, const GData & arg0, const GData & arg1, 
		   const GData & arg2,const GData & arg3, const GData & arg4, 
		   const GData & arg5, const GData & arg6) {
		      type = GDT_Term ;
		      val.term = NewTerm(symval(sym), arg0, arg1, arg2, arg3, 
					 arg4, arg5, arg6) ;
      } ;
      inline GTerm(const GData& sym, const GData & arg0, const GData & arg1, 
		   const GData & arg2, const GData & arg3, const GData & arg4,
		   const GData & arg5, const GData & arg6, const GData & arg7){
		      type = GDT_Term ;
		      val.term = NewTerm(symval(sym), arg0, arg1, arg2, arg3, 
					 arg4, arg5, arg6, arg7) ;
      } ;

      //--------------------------------------------------
      // describe

      void _describe(ostream& o,const bool newlinep,const bool detailp) 
   const ;
//  virtual void __describe(ostream& o,const bool newlinep,const bool detailp) 
//    const ;
} ;

//======================================================================
// check data type
//
inline bool isGInteger(const GData &dt) { return dt.type == GDT_Integer ; } ;
inline bool isGFloat(const GData &dt) { return dt.type == GDT_Float ; } ;
inline bool isGSymbol(const GData &dt) { return dt.type == GDT_Symbol ; } ;
inline bool isGVariable(const GData &dt) { return dt.type == GDT_Variable ; } ;
inline bool isGString(const GData &dt) { return dt.type == GDT_String ; } ;
inline bool isGPointer(const GData &dt) { return dt.type == GDT_Pointer ; } ;
inline bool isGTerm(const GData &dt) { return dt.type == GDT_Term ; } ;

inline bool isGNumber(const GData &dt) { 
   return isGInteger(dt)||isGFloat(dt) ; 
} ;

inline bool isGAtom(const GData &dt) {
   return  isGSymbol(dt) || isGNumber(dt) || isGString(dt) || isGPointer(dt) ;
} ;

inline bool eq(const GData& x, const GData& y) { return x.eq(y) ; };

inline Int IntVal(const GData &dt) {
   if(isGInteger(dt)) return dt.val.inum ;
   if(isGFloat(dt)) return (Int)(dt.val.fnum) ;
   return 0 ;
} ;
inline Float FltVal(const GData &dt) {
   if(isGFloat(dt)) return dt.val.fnum ;
   if(isGInteger(dt)) return (Float)(dt.val.inum) ;
   return 0.0 ;
} ;

inline Symbol* _SymVal(const GData& dt) { return dt.val.symbol ; } ;
inline Symbol* _VarVal(const GData& dt) { return dt.val.symbol ; } ;
inline Symbol* SymVal(const GData& dt) {
   if(isGSymbol(dt)) return _SymVal(dt) ;
   else return _SymVal(NullSymbol) ;
} ;

inline GCell* _SymGCell(const GData& dt) { return _SymVal(dt)->gcell ; } ;
inline Gem* _SymGem(const GData& dt) { return _SymVal(dt)->gem ; } ;
inline GDataBox* _SymGDBox(const GData& dt) { return _SymVal(dt)->gdbox ; } ;
inline GMutex* _SymMutex(const GData& dt) { return _SymVal(dt)->mutex ; } ;

inline bool hasGCell(const GData& dt) { 
   return isGSymbol(dt) && _SymVal(dt)->hasGCell() ; 
}
inline bool hasGem(const GData& dt) { 
   return isGSymbol(dt) && _SymVal(dt)->hasGem() ; 
}
inline bool hasGDBox(const GData& dt) { 
   return isGSymbol(dt) && _SymVal(dt)->hasGDBox() ; 
}
inline bool hasMutex(const GData& dt) { 
   return isGSymbol(dt) && _SymVal(dt)->hasMutex() ; 
}

inline Term* _TermVal(const GData& dt) { return dt.val.term ; } ;
inline Symbol* _TermHead(const GData& dt) { return dt.val.term->head ; } ;
inline Int _TermArity(const GData& dt) { return dt.val.term->arity ; } ;
inline GData& _TermArgv(const GData& dt, Int n) { 
   return dt.val.term->argv(n) ; } ;

inline String* _StrVal(const GData& dt) { return dt.val.string ; } ;
inline void* _PtrVal(const GData& dt) { return dt.val.pointer ; } ;

//======================================================================
// GData Box
//

class GDataBox {
   public:
      GData val ;
      bool assigned ;
      CondBarrier mutex ;

      //--------------------------------------------------
      // Creater

      inline GDataBox() {
	 val = NullSymbol ;
	 assigned = false ;
      } ;

      inline GDataBox(const GDataBox& src) {
	 val = src.val ;
	 assigned = src.assigned ;
      } ;

      inline GDataBox(const GData& v) {
	 val = v ;
	 assigned = true ;
      } ;

      //--------------------------------------------------
      // put

      enum PutType {
	 PT_nowait,
	 PT_wait,
	 PT_force
      } ;

      bool put(const GData& v, PutType puttype) ;

      //--------------------------------------------------
      // get

      bool get(GData& v, bool retractp, bool waitp) ;

      //--------------------------------------------------
      // Printer
  
      inline void describe(ostream& o, bool detailp = true) {
	 o << "#GDataBox[" << assigned << ":" << val << "]" ;
	 if(detailp) o << endl ;
      } ;
      inline friend ostream& operator<< (ostream& o, GDataBox& gdbox) {
	 gdbox.describe(o,false) ;
	 return o ; 
      } ;
      
} ;



//======================================================================
// Special Data
//

#include "gglobal.h"

#define VARCONCATCHAR "_"

inline bool isNullSymbol(GData& dt) { return eq(dt,NullSymbol) ; } ;
inline bool isConsTerm(GData& dt) { 
   return (isGTerm(dt) &&
	   _TermHead(dt) == _SymVal(ConsSymbol) &&
	   _TermArity(dt) == 2) ; 
} ;
inline GData _ConsCar(GData& dt) { return _TermArgv(dt,0) ; } ;
inline GData _ConsCdr(GData& dt) { return _TermArgv(dt,1) ; } ;

inline void GData::init() { *this = NullSymbol ; } ;

GData genGSymbol(const GData &seed) ;
inline GData genGSymbol() { return genGSymbol(NullSymbol) ; } ;

GData genGVariable(const GData &seed) ;
inline GData genGVariable() { return genGVariable(AnonymousVar) ; } ;

//======================================================================
#include "gcell.h"

////////////////////////////////////////////////////////////////////////
#endif


