// -*- Mode:C++ -*-
//Header:
//File: gem.h
//Title: gaea elementary machine
//Author: NODA Itsuki
//Date: 1998/6/09
//EndHeader:

//ModifyHistory:
// 1998/6/09: Create this file
//EndModifyHistory:

//Memo:
//EndMemo:

#ifndef _GEM_H_
#define _GEM_H_
////////////////////////////////////////////////////////////////////////

#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>
#include "thread_util.h"
#include "mmtoolT.h"
#include "gdata.h"
#include "gbio.h"
#include "unify.h"
#include "gcell.h"

//======================================================================
// AndState
//

class AndState {
   public:
      InstGData igdata ;
      UInt cutOrLevel ;
      GClausePointer gclpointer ;
      
      inline AndState() {
	 init() ;
      } ;
      inline AndState(const AndState& src) {
	 igdata = src.igdata ;
	 cutOrLevel= src.cutOrLevel ;
	 gclpointer = src.gclpointer ;
      } ;
      inline AndState(const GData& goal, Binding *binding, UInt cutlevel = 0) {
	 init(goal,binding,cutlevel) ;
      } ;
      inline AndState(const InstGData& goal, UInt cutlevel = 0) {
	 init(goal,cutlevel) ;
      } ; 

      inline AndState operator=(const AndState& src) {
	 igdata = src.igdata ;
	 cutOrLevel= src.cutOrLevel ;
	 gclpointer = src.gclpointer ;
	 return *this ;
      } ;

      inline void init() {
	 igdata = InstGData() ;
	 cutOrLevel = 0 ;
	 gclpointer.init() ;
      } ;

      inline void init(GData goal, Binding* binding, UInt cutlevel = 0) {
	 init(InstGData(goal,binding), cutlevel) ;
      } ;
      inline void init(const InstGData& goal, UInt cutlevel = 0) {
	 igdata = goal ;
	 cutOrLevel = cutlevel ;
	 gclpointer.init() ;
      } ;

      inline void initGClausePointer(GCell* gcell) {
	 gclpointer.init(gcell) ;
      } ;

      inline bool firsttryp() {
	 return isNullPtr(gclpointer.top) ;
      } ;
	 
      inline void describe(ostream& o, bool detailp = true) const {
	 o << "#AndState[" << igdata ;
	 o << " ,cutLevel=" << cutOrLevel ;
	 o << " ,gcp=" << gclpointer ;
	 if(detailp) { o << endl ; } ;
      } ;
      inline friend ostream& operator<<(ostream& o, const AndState& astate) {
	 astate.describe(o,false) ;
	 return o ;
      } ;
} ;

typedef SpStack<AndState> AndStateStack ;
typedef SpStackEntry<AndState> AndStateStackEntry ;

//======================================================================
// OrState
//

class OrState {
   public:
      AndStateStack::rewindPos andStateStackPointer ;
      BindingState bindingState ;
      UInt gclausePointerRewindPos ;

      //--------------------------------------------------
      // Constructor

      inline OrState() {
	 andStateStackPointer = AndStateStack::rewindPos(NULLPTR,0) ;
	 gclausePointerRewindPos = 0 ;
      } ;

      inline OrState(const OrState& src) {
	 andStateStackPointer = src.andStateStackPointer ;
	 bindingState = src.bindingState ;
      } ;
      inline OrState(AndStateStack::rewindPos assp,
		     BindingState bstate)
      {
	 andStateStackPointer = assp ;
	 bindingState = bstate ;
      } ;

      //--------------------------------------------------
      // Let

      inline OrState operator=(const OrState& src) {
	 andStateStackPointer = src.andStateStackPointer ;
	 bindingState = src.bindingState ;
	 return *this ;
      } ;

      //--------------------------------------------------
      // output

      void describe(ostream& o, bool detailp = true) const ;

      inline friend ostream& operator<< (ostream& o, const OrState& os) {
	 os.describe(o,false) ;
	 return o ;
      } ;

} ;

typedef AStack<OrState> OrStateStack ;

//======================================================================
// GEM Gaea Elementary Machine
//

typedef AHeap<GClausePointer> GClausePointerHeap ;

#define DEFAULTGEMPREFIX	"gem"
#define ANDSTATESTACKINITSIZE	128
#define ORSTATESTACKINITSIZE	128
#define UIHEAPINITSIZE		256
#define GCPHEAPINITSIZE		512

class Gem {
   public:
      pthread_t pid ;
      Symbol* name ;
      GCell* gcell ;

      inline GCell* agcell() {
	 if(gcell->gcellStack.bottomp()) return gcell ;
	 else return gcell->gcellStack.top() ;
      } ;

      InstGData topGoal ;

      AndState* astate ;

      inline InstGData& currentGoal() const {
	 return astate->igdata ;
      } ;
      inline GClausePointer* gclausePointer() const {
	 return &(astate->gclpointer) ;
      } ;

      enum TryStatus { TS_none, TS_call, TS_success, TS_fail, TS_match } ;
      TryStatus tstate ;

      AndStateStack andStateStack ;
      OrStateStack orStateStack ;

      UnifyInfoHeap uiheap ;
      
      enum traceLevel { TRC_none, TRC_normal, TRC_kernel, TRC_max } ;
      traceLevel tracelevel ;

      inline void traceon(traceLevel tl) { tracelevel = tl ; } ;
      inline void notrace() { tracelevel = TRC_none ; } ;

      //--------------------------------------------------
      // Constructor

      inline Gem() { init() ; } ;
      inline Gem(Symbol* nm, bool newsymbolp = true) {
	 init(nm, newsymbolp) ;
      } ;
      inline Gem(const char* nm, bool newsymbolp = true) {
	 init(nm,newsymbolp) ;
      } ;

      inline void init(Symbol* nm = NULLPTR, bool newsymbolp = true) {
	 if(isNullPtr(nm)) {
	    name = GenSymbol(DEFAULTGEMPREFIX) ;
	 } else if (newsymbolp) {
	    name = GenSymbol(nm) ;
	 } else {
	    name = nm ;
	 } ;
	 
	 pid = pthread_self() ;
	 
	 gcell = new GCell(name) ;
	 name->gem = this ;
	 gcell->pushGCell(BasicCell) ;
	 gcell->pushGCell(MainCell) ;

	 topGoal = InstGData() ;

	 astate = NULLPTR ;

	 andStateStack.expand(ANDSTATESTACKINITSIZE,false) ;
	 orStateStack.expand(ORSTATESTACKINITSIZE,false) ;
	 uiheap.init(UIHEAPINITSIZE,false) ;
	 tracelevel = TRC_none ;
      } ;

      inline void init(const char* nm, bool newsymbolp = true) {
	 init(FindSymbol(nm),newsymbolp) ;
      } ;

      //--------------------------------------------------
      // And State
      
      inline AndState* pushAndState(const InstGData& newgoal,
					      UInt cutlevel) 
      {
//	 return andStateStack.push(AndState(newgoal,cutlevel)) ;
	 AndState* as = andStateStack.pushget() ;
	 as->init(newgoal,cutlevel) ;
	 return as ;
      } ;
      inline AndState* pushAndState(const InstGData& newgoal) {
	 return pushAndState(newgoal,0) ;
      } ;
      inline AndState* pushAndState(const GData dt, 
					      const Binding* binding) {
	 return pushAndState(InstGData(dt,binding)) ;
      } ;
      inline AndState* pushAndState(const GData dt, 
				    const Binding* binding,
				    UInt cutlevel)
      {
	 return pushAndState(InstGData(dt,binding),cutlevel) ;
      } ;
      
      inline AndState& popAndState() {
	 AndState* as = &(andStateStack.pop()) ;
	 if(!andStateStack.bottomp()) 
	    andStateStack.top().gclpointer.init() ;
	 return *as ;
      } ;

      inline UInt topposAndStateStack() {
	 return andStateStack.toppos() ;
      } ;

      inline UInt rewindAndStateStack(UInt pos) {
	 return andStateStack._top = pos ;
      } ;

      //--------------------------------------------------
      // Or State

      inline OrState* pushOrState() {
	 orStateStack.push(OrState(andStateStack.rewindpos(),
				   memoryBindingState(currentGoal().binding,
						      &uiheap))) ;
	 return &(orStateStack.top()) ;
      } ;
      inline void popOrState() {
	 OrState ostate = orStateStack.pop() ;
	 andStateStack.rewind(ostate.andStateStackPointer) ;
	 rewindBindingState(ostate.bindingState) ;
      } ;

      //--------------------------------------------------
      // Unfold

      enum UnfoldState {
	 US_fail,
	 US_continue,
	 US_succeed,
      } ;

      void load(String filename, GCell* dstcell = NULLPTR) ;
      inline void load(char* filename, GCell* dstcell = NULLPTR) { 
	 load(String(filename), dstcell) ; 
      } ;
      void loadloop(istream& istr, GCell* dstcell = NULLPTR) ;

      void reploop0(istream& istr = cin, ostream& ostr = cout) ;
      void settopgoal(GData goal, Binding* binding = NULLPTR, 
		      bool resetp = true) ;
      bool unfoldLoop() ;
      UnfoldState unfold() ;
      bool tryDef() ;
      bool trySystemDef() ;
      bool tryUserDef() ;

      //--------------------------------------------------
      // Cell Operation

      inline bool pushGCell(GCell* gc, GCell* dst) { 
	 dst->pushGCell(gc) ; 
	 return true ;
      } ;
      inline bool pushGCell(GCell* gc) { return pushGCell(gc, gcell) ;} ;
      inline bool pushGCell(GData gc, GData src) { 
	 if(isGSymbol(gc) && isGSymbol(src))
	    return pushGCell(_SymVal(gc)->gcell,_SymVal(src)->gcell) ;
	 return false ;
      } ;
      inline bool pushGCell(GData gc) {
	 if(isGSymbol(gc))
	    return pushGCell(_SymVal(gc)->gcell) ;
	 return false ;
      } ;

      //--------------------------------------------------
      // Output
      
      void describe(ostream& o, bool detailp = true) const ;
      void showstatus(ostream& o) const ;

      inline friend ostream& operator<<(ostream& o, const Gem& gem) {
	 gem.describe(o,false) ;
	 return o ;
      } ;

} ;

extern void dbdGem(Gem* gem) ;
extern void* UnfoldLoop(void* gem) ;
     
////////////////////////////////////////////////////////////////////////
#endif
