// -*- Mode:C++ -*-
//Header:
//File: ring.h
//Author: NODA Itsuki
//Date: 1998/2/9
//EndHeader:

//ModifyHistory:
// 1998/2/9: Create this file
// 1998/5/27: reconstruct 
//EndModifyHistory:

//Memo:
//	typedef Ring<ContentType> FOO ;
//	FOO foo1(data1) ;
//	FOO foo2(data2) ;
//	foo1.jointLast(&foo2) ;
//	foo1.reverse() ;
//	foo1.split(&foo2) ;
//	foo1.describe(cout) ;
//EndMemo:

#ifndef _ring_h_
#define _ring_h_
////////////////////////////////////////////////////////////////////////

//======================================================================
// Ring<Content>
//

template <class _Contents_>
class Ring {
   public:
      //--------------------------------------------------
      // LocalTypeDef
      typedef Ring<_Contents_> ring_t ; 
      typedef _Contents_ cont_t ;

      //--------------------------------------------------
      // Slots
      cont_t entity ;
      ring_t* previous ;
      ring_t* next ;

      //--------------------------------------------------
      // Constructor
      inline Ring<_Contents_> () {
	 previous = NULLPTR ;
	 next = NULLPTR ;
      } ;
      Ring<_Contents_> (const cont_t val, bool circlep = false) {
	 entity = val ;
	 if(circlep) {
	    previous = this ;
	    next = this ;
	 } else {
	    previous = NULLPTR ;
	    next = NULLPTR ;
	 }
      } ;

      //--------------------------------------------------
      // set data

      ring_t* set(const cont_t val) {
	 entity = val ;
      } ;

      //--------------------------------------------------
      // Add Next/Previous
      ring_t* addNext(ring_t* newnext) {
	 ring_t* oldnext = next ;
	 
	 newnext->previous = this ;
	 newnext->next = oldnext ;
	 next = newnext ;
	 if(oldnext != NULLPTR) oldnext->previous = newnext ;

	 return this ;
      } ;
      
      ring_t* addPrevious(ring_t* newprev) {
	 ring_t* oldprev = previous ;
	 
	 newprev->previous = oldprev ;
	 newprev->next = this ;
	 previous = newprev ;
	 if(oldprev != NULLPTR) oldprev->next = newprev ;

	 return this ;
      } ;
      
      //--------------------------------------------------
      // Insert new data (allocate new memory)
      ring_t* insertFirst(const cont_t& dt) {
	 ring_t* newring = new ring_t(dt) ;
	 addNext(newring) ;
	 return newring ;
      } ;

      ring_t* insertLast(const cont_t& dt) {
	 ring_t* newring = new ring_t(dt) ;
	 addPrevious(newring) ;
	 return newring ;
      } ;

      //--------------------------------------------------
      // Remove Entry
      ring_t* removeSelf() {
	 ring_t* pre = previous ;
	 ring_t* nxt = next ;
	 pre->next = next ;
	 nxt->previous = pre ;
	 return this ;
      } ;

      ring_t* remove(ring_t* del) {
	 return del->removeSelf() ;
      } ;

      //--------------------------------------------------
      // join Next
      ring_t* joinNext(ring_t* newnext) {
	 ringlet_t* oldnext = next ;
	 ringlet_t* oldnprev = newnext->previous ;

	 next = newnext ;
	 oldnprev->next = oldnext ;
	 newnext->previous = this ;
	 oldnext->previous = oldnprev ;

	 return this ;
      } ;

      //--------------------------------------------------
      // Join Previous
      ring_t* joinPrevious(ring_t* newprev) {
	 ring_t* oldprev = previous ;
	 ring_t* oldnprev = newprev->previous ;
	  
	 previous = oldnprev ;
	 newprev->previous = oldprev ;
	 oldprev->next = newprev ;
	 oldnprev->next = this ;

	 return this ;
      } ;

      //--------------------------------------------------
      // Split
      ring_t* split(ring_t* tail) {
	 ring_t* cprev = previous ;
	 ring_t* tprev = tail->previous ;

	 cprev->next = tail ;
	 tail->last = cprev ;
	 previous = tprev ;
	 tprev->next = this ;

	 return this ;
      } ;

      //--------------------------------------------------
      // Reverse
      ring_t* reverse() {
	 ring_t* r = this ;
	 ring_t* nr ;
	 do {
	    nr = r->next ;
	    r->next = r->previous ;
	    r->previous = nr ;
	    r = nr ;
	 } while(r != this) ;
	 return this ;
      } ;

      //--------------------------------------------------
      // Single Ring Check
      
      inline bool isSingleRing() {
	 return next == this ;
      } ;

      //--------------------------------------------------
      // Length

      Int length() {
	 ring_t* r = this ;
	 Int l = 0 ;
	 do {
	    r = r->next ;
	    l++ ;
	 } while(r != this) ;
	 return l ;
      } ;

      // length of circle ring entity
      inline Int clength() { return length() - 1 ; } ;
      
      //--------------------------------------------------
      // Describe
      void describe(ostream& o, bool detailp = false) const {
	 if(detailp && !isNullPtr(this)) {
	    o << "#Ring[" << this << "]" ;
	    ring_t* r = (ring_t*)this ;
	    do {
	       o << endl << "\t" << r << "\t" << r->entity ;
	       r = r->next ;
	    } while (r != this) ;
	    o << endl << "\t------" << endl ;
	 } else {
	    o << "#Ring[" << this ;
	    if(!isNullPtr(this)) {
	       o << ":cont=" << entity 
		 << ", prev=" << previous << ", next=" << next ;
	    }
	    o << "]" ;
	    if(detailp) o << endl ;
	 }
      } ;
      
      //--------------------------------------------------
      // Output
      friend ostream& operator<< (ostream& o, const ring_t& r) {
	 r.describe(o) ;
	 return o ;
      } ;
} ;

////////////////////////////////////////////////////////////////////////
#endif
